\name{runHmmld.ped}
\alias{runHmmld.ped}
\title{A function for estimating relatedness between to individuals
  using SNP data, This function should only be used if you have a
  'snpMatrix object' and corrosponding 'subject.support' and 'snp.support'}
\description{
  This function estimates the probability of sharing alleles identity by
  descent (IBD) accross the genome. The method is based on a continouos
  time Markov model with hidden states. The hidden states are the IBD
  states between two diploid chromosome pairs. We assume that the
  individuals are not inbreed and thus the individuals can share 0, 1 or
  2 alleles IBD. The SNPs are allowed to be in linkage disequilibrium
  (LD). To accomidate LD the methods need SNP for several individuas in
  order to estimate the allele frequencies and the pairwise LD. The
  method return the posterior probabilies of the IBD states accross the
  genome  and the overall IBD sharing. This function will use a
  snpMatrix object consisting of chromosome, position and genotypes.
  }
\usage{
runHmmld.ped(pedfile_object,pair=c(1,2),par=NULL,min.maf=0,LD="rsq2",epsilon=0.01,back=5,alim=c(0.001,0.15),start=NULL,prune=NULL,ld_adj=TRUE,fix.a=NULL,fix.k2=NULL,calc.a=FALSE,phi=0.013,timesToRun=10,timesToConverge=5,giveCrap=FALSE,convTol=0.1)
}

\arguments{
  \item{pedfile_object}{Return value from 'SnpMatrix' function
  'read.snps.pedfile'. This is a list of 3 elements. 'snp.matrix,
  'subject.support' and 'snp.support. The  matrix should be SNP genotypes where NA or 0
  denotes missing data, 1 for AA, 2 for Aa  and 3 for aa. The number of
  individuals is the number of rows and the number of SNPs is the number
  of coloums. The position of each SNP in centi Morgan (or mega
  bases). If centi Morgan is used then phi should be set to 1, a vector
  of chromosomes numbers. }
  \item{pair}{Integer vector of length two with the row numbers of the two individuals where relatedness is to be estimated}
  \item{par}{Optional numeric vector c(a,k2,k1,k0) of parameters used instead of optimazation}
  \item{min.maf}{The minumum minor allele frequecy allowed}
  \item{LD}{The measure used to select the previous SNP to condition on, ("D'","D" or "rsq2")}
  \item{epsilon}{The error rate}
  \item{back}{The number of privous SNPs that can be conditioned on (see details for recomadations) }
  \item{alim}{The allowed range for a}
  \item{start}{Optional starting point for the optimazition}
  \item{prune}{The maximum value allowed for pairwise LD. If 0 then to pruning is performed}
  \item{ld_adj}{Logical. use the pairwise emission probabilities to correct for LD}
  \item{fix.a}{Numeric. Fix the a parameters to this value}
  \item{fix.k2}{Numeric. Fix the k2 parameter to this value}
  \item{calc.a}{Estimate the a parameter from the overall IBD sharing. appropreate for distantly related individuals or individual who are related though one or two paths of the same length}
  \item{phi}{Numeric. The recombination rate in Morgans per Mega base (m/Mb)}
  \item{timesToRun}{Integer. The maximum number of times the optimization is run}
  \item{timesToConverge}{Integer. The number of times the optimazation should reach the same optimum}
  \item{giveCrap}{Integer. If non zero then emission probabilies (given
  the unobserved state and allele phase) and the haplotype probabilities
  are returned. Also more runtime information is given.}
  \item{convTol}{Numeric. The tolerance for stating that the likelihood have reached the same likelihood.}
}
\details{
How to select the number of privious SNP that can be conditioned
  on. First of all if is no LD in the data chosse back=0.If you want to
  prune SNP away based on LD and/or you want to accomidate LD in the
  model choose <back> as the number of SNP where you expect there to be
  LD. For example if you have 500,000 SNP you expect their to be LD
  between a lot of SNPs in a region. Here I use back=50. If you only
  have 10,000 SNPs I woulduse back=5. 

  If there is LD in the data but you want to remove the LD before the
  analysis then set adj\_ld to FALSE, prune to some numeric value larger
  than zero (e.g. 0.2) and back to some number.
  
}
\value{
  \item{kResult}{The value for sharing 2, 1 or 0 alleles IBD}
  \item{kLike}{The maximal -log likelihood}
  \item{kr}{The co-ancestry coefficient}
  \item{a}{The a parameter}
  \item{uLike}{The likehood for being unrelated}
  \item{LD}{The LD measure}
  \item{t}{distance between the used SNPs (some SNPs might have been discarded)}
  \item{snp}{The number of used SNPs (some SNPs might have been discarded)}
  \item{position}{The position of the used SNPs  (some SNPs might have been discarded)}
  \item{double_recom}{If false then no instanstationus double recombination is allowed}
  \item{alim}{the allowed range of a}
  \item{poLike}{The likelihood for being parent-offspring}
  \item{back}{The number of privious SNPs that have been conditioned on (and/or used for pruning)}
  \item{post}{a matrix with posterior probabilities for the hidden IBD states}
  \item{timesRun}{The number of times the optimazation algorihm have been run}
  \item{timesConverged}{The number of times the optimazation algorithm have found the same maximum}
  \item{convergenceInfo}{matrix with estimates of the overall IBD sharing and the likelihood for each of the optimaziations}
  \item{usedSnps}{a vector indicating wether a SNP have been used in the analysis (1's) or discarded (0's)}
  \item{S}{The joint emission probabilies for the current SNP and the privious SNP}
  \item{choose}{vector indivicating which privious SNP have been used to condition on}
  \item{mea}{The LD meassure for the <back> number of privious SNPs}
  \item{S1}{The emission probabilies for each single SNP}
  \item{hap}{The haplotypes probabilies for each privious SNPs}
  \item{maf}{the minor allele frequency for the SNPs}
	\item{path}{The famous viterbi path}
}
\references{http://staff.pubhealth.ku.dk/~ande/web/software/relate.html }
\author{Anders Albrechtsen }

\seealso{ \code{snpMatrix} and \code{\link[snpMatrix]{read.snps.pedfile}} }
\examples{
##
##  SNP MATRIX MUST BE INSTALLED
# library("snpMatrix")
#path<-paste(.find.package("Relate"),"/data/500.ped",sep="")
#ped<-read.snps.pedfile(path)
#pedRes<-runHmmld.ped(ped)
#plot(pedRes)
}
\keyword{models}
